import networkx as nx
import numpy as np
import higra as hg

class Graph:
    def __init__(self, is_binary, hierarchy):
        self.is_binary = is_binary
        if hierarchy in ['watershed_hierarchy_by_attribute', 
                      'watershed_hierarchy_by_minima_ordering', 
                      'watershed_hierarchy_by_volume', 
                      'watershed_hierarchy_by_area'
                      'watershed_hierarchy_by_dynamics', 
                      'watershed_hierarchy_by_number_of_parents']:
          self.hierarchy = hierarchy
        else: 
          self.hierarchy = 'watershed_hierarchy_by_area'

    def cut_graph(self, file, cut_graph_file, cutNumber):
      RG = nx.Graph()
      cut_list = []

      with open(file.file) as f:
        lines = f.readlines()

      line = len(lines)-1
      cut_list.append(int(lines[line].split(", ")[0]))

      while(line >= 0):
        cut = lines[line]
        v1 =int(cut.split(", ")[0]) # node 1
        v2 =int(cut.split(", ")[1]) # node 2
        w = float(cut.split(", ")[2]) # weight

        if (v2 in cut_list):
          if cutNumber >0:
            cutNumber -= 1
            cut_list.append(v1)
          else:
            cut_graph_file.save_graph_data(v1, v2, w)
            RG.add_node(v1)
        else:
          cut_graph_file.save_graph_data(v1, v2, w)
          RG.add_edge(v1, v2, weight = w)
        line -=1

      return RG

    def compute_hierarchy(self, input_g, input_higra):
      leaf_list = []
      graph = hg.UndirectedGraph()       #convert the scikit image rag to higra unidrect graph
      graph.add_vertices(max(input_g._node)+1)   #creating the nodes (scikit image RAG starts from 1)
      edge_list = list(input_g.edges())                   #ScikitRAG edges
      size_threshold = 20

      for i in range (len(edge_list)):
          graph.add_edge(edge_list[i][0], edge_list[i][1]) #Adding the nodes to higra graph
      edge_weights = np.empty(shape=len(edge_list))
      sources, targets = graph.edge_list()

      for i in range (len(sources)):
        edge_weights[i] = int(input_g.adj[sources[i]][targets[i]]["weight"])

      if self.hierarchy == 'watershed_hierarchy_by_attribute':
        nb_tree, nb_altitudes = hg.watershed_hierarchy_by_attribute(graph, edge_weights)
      elif self.hierarchy == 'watershed_hierarchy_by_minima_ordering':
        nb_tree, nb_altitudes = hg.watershed_hierarchy_by_minima_ordering(graph, edge_weights)
      elif self.hierarchy == 'watershed_hierarchy_by_volume':
        nb_tree, nb_altitudes = hg.watershed_hierarchy_by_volume(graph, edge_weights)
      elif self.hierarchy == 'watershed_hierarchy_by_area':
        nb_tree, nb_altitudes = hg.watershed_hierarchy_by_area(graph, edge_weights)
      elif self.hierarchy == 'watershed_hierarchy_by_dynamics':
        nb_tree, nb_altitudes = hg.watershed_hierarchy_by_dynamics(graph, edge_weights)
      elif self.hierarchy == 'watershed_hierarchy_by_number_of_parents':
        nb_tree, nb_altitudes = hg.watershed_hierarchy_by_number_of_parents(graph, edge_weights)

      if(self.is_binary):
        tree, node_map = hg.tree_2_binary_tree(nb_tree)
        altitudes = nb_altitudes[node_map]
      else:
        tree = nb_tree
        altitudes = nb_altitudes

      for n in tree.leaves_to_root_iterator():
        leaf = -2 # It's cod is used for the node that is not a leaf
        if(tree.is_leaf(n)):
          leaf = -1 # It's cod is used for the node that is a leaf
          leaf_list.append(n)
        input_higra.save_graph_data(n, tree.parent(n), leaf)

      return(leaf_list)


import os
import numpy as np
import tensorflow as tf
from tensorflow.keras.applications.vgg16 import VGG16, preprocess_input
from tensorflow.keras.applications.resnet50 import ResNet50
from tensorflow.keras.models import Model

class Models:
  def __init__(self, model_name):
    self.available_models = ['vgg16', 'resnet50']
    self.include_top = True

    # verify if the model is valid
    if model_name in self.available_models:
      try:
        if model_name == 'vgg16':
          self.name = model_name
          self.layername_feature_extraction = 'fc2'
          self.img_size_model = (224, 224)
          self.model = VGG16(weights='imagenet', include_top=self.include_top)
        elif model_name == 'resnet50':
          self.name = model_name
          self.layername_feature_extraction = 'predictions'
          self.img_size_model = (224, 224)
          self.model = ResNet50(weights='imagenet', include_top=self.include_top)
      except:
        print(f">> Error while loading model '{model_name}'")
    else:
        print(f">> Error: there is no '{model_name}' in {self.available_models}")

  def features(self, video_file):
    features_list = []
    if(os.path.exists(video_file)):
      frame_list = os.listdir(video_file)
      frame_list.sort() # to garanted the time order
      features_list = [    # Compute feature vector extracted
                      self.get_feature_vector(video_file + frames) for frames in frame_list #List of frame features
                      ]

    return features_list

  def get_feature_vector(self, img_path):
      model_feature_vect = Model(inputs=self.model.input, outputs=self.model.get_layer(self.layername_feature_extraction).output)
      img = tf.keras.utils.load_img(img_path, target_size=self.img_size_model)
      img_arr = np.array(img)
      img_ = self.image_processing(img_arr)

      feature_vect = model_feature_vect.predict(img_)

      return feature_vect

  def image_processing(self, img_array):
      img = np.expand_dims(img_array, axis=0)
      processed_img = preprocess_input(img)

      return processed_img


from .Files import Files
from .Frame import Frame
from .Models import Models
from .Graph import Graph
import os 
from PIL import Image
import numpy as np
from scipy import spatial
import networkx as nx
import cv2 as cv

class Summary:
  def __init__(self, dataset_frames, video, rate, time, hierarchy, selected_model, is_binary, percent, alpha, keyshot, keyframe):
    self.delta_t = rate * time
    self.video_file = "{}/{}/".format(dataset_frames, video)
    self.video = video
    self.alpha = alpha
    self.percent = percent
    self.graph = Graph(is_binary, hierarchy)
    self.model = Models(selected_model)
    self.frame = Frame(self.model)
    self.len_shot = 0
    self.rate = rate

    # creating path for files
    self.input_graph_file = Files('{}graph.txt'.format(self.video_file))
    if keyshot: 
      self.input_keyshot = Files('{}keyshot_{}_{}.txt'.format(self.video_file, self.percent, self.alpha))
      self.input_keyshot_path = '{}keyshot_{}_{}/'.format(self.video_file, self.percent, self.alpha)
    self.input_mst = '{}mst_{}_{}.txt'.format(self.video_file, self.percent, self.alpha)
    self.input_higra = Files('{}higra_{}_{}.txt'.format(self.video_file, self.percent, self.alpha))
    self.cut_graph_file = Files('{}cut_graph_{}_{}.txt'.format(self.video_file, self.percent, self.alpha))
    self.frames_path = "{}frames/".format(self.video_file)
    self.output_skim = '{}skim_{}_{}'.format(self.video_file, self.percent, self.alpha)
    if keyframe:
      self.input_keyframe_path = '{}keyframe_{}_{}/'.format(self.video_file, self.percent, self.alpha)
      self.input_keyframe = Files('{}keyframe_{}_{}.txt'.format(self.video_file, self.percent, self.alpha))

    cut_number = self.bestCutNumber()

    print("----------------------")
    print("Processing video {}".format(self.video_file))

    if(not os.path.exists(self.input_graph_file.file)):
      f = open(self.input_graph_file.file, "a") # pensando em paralelizar para garantir a integridade do arquivo
      f.close()
      features_list = self.model.features(self.frames_path) # extract features
      self.frame.load(self.frames_path, self.delta_t, self.input_graph_file, features_list) # Load the frame list and create a graph for the video

    if(not os.path.exists(self.input_higra.file)):

      self.len_shot = round((len(os.listdir(self.frames_path))) * (self.percent/100)) - 1
      self.cut_number = int(self.bestCutNumber() * alpha)
      if(self.cut_number <= 2):
          self.cut_number = 3

      #tree = gen_mst(input_graph_file, input_mst) # generate the minimum spanning tree
      tree = self.input_graph_file.read_graph_file(Files(), cut_graph = False, cut_number = 0) # read the graph file
      leaflist = self.graph.compute_hierarchy(tree, self.input_higra) # Create the hierarchy based on the minimum spanning tree and return the leaves of the new hierarchy
      cuted_graph = self.graph.cut_graph(self.input_higra, self.cut_graph_file, cutNumber = cut_number) # Create a new graph based on the hierarchy and the level cut
      #plotGraph(cuted_graph, True)
      if keyframe:
        self.selectKeyFrame(cuted_graph, leaflist) # With the cuted graph, create a keyframe to represent each component or segment of video
      if keyshot:
        self.selectKeyShot(cuted_graph, leaflist)
      if(not os.path.exists(self.output_skim)):
        os.mkdir(self.output_skim)
      self.generate_video()


  def bestCutNumber(self):
    if(os.path.exists(self.frames_path)):
        frame_list = os.listdir(self.frames_path)
        frame_list.sort() # to garanted the time order
        features_list = []
        for frames in frame_list:
            if frames.endswith("jpg"):
                frame_dir = self.frames_path + frames
                features_list.append(self.rgbSim(frame_dir))
        # features_list = [rgbSim(video_file + frames) for frames in frame_list]
        weight_list = []
        feat_list_len = len(features_list)

        for vertex1 in range(feat_list_len):
            for vertex2 in range(self.calc_init(vertex1, self.delta_t, feat_list_len),
                                    self.calc_end(vertex1, self.delta_t, feat_list_len)):
                w = self.spatialSim(features_list[vertex1], features_list[vertex2])/100 # teste nan
                weight_list.append(w)
        cut = np.std(weight_list)
        while cut<=1:
            cut = cut * 10
        return(np.round(cut) - 1)

  def rgbSim(self, frame_dir):
      frame = Image.open(frame_dir)
      frame_reshape = frame.resize((round(frame.size[0]*0.5), round(frame.size[1]*0.5)))
      frame_array = np.array(frame_reshape)
      frame_array = frame_array.flatten()
      frame_array = frame_array/255
      return frame_array

  def spatialSim(self, frame1, frame2):
      similarity = 100 * (-1 * (spatial.distance.cosine(frame1, frame2) - 1))
      if similarity < 20:
          similarity = 20
      return similarity

  def selectKeyFrame(self, graph, leaflist):
    S = [graph.subgraph(c).copy() for c in nx.connected_components(graph)]
    #KF_list = []
    for c in range(len(S)):
      central_node = len(S[c].nodes)
      comp_leaf_list = []
      for i in range(central_node):
        if(list(S[c])[i] in leaflist):
          comp_leaf_list.append(list(S[c])[i])
      cn = int(len(comp_leaf_list)/2)
      if not (cn == 0):
        kf = str(comp_leaf_list[cn]).zfill(6)

        if not os.path.isdir(self.input_keyframe_path):
          os.mkdir(self.input_keyframe_path)
        os.system('cp {}frames/{}.jpg {}{}.jpg'.format(self.video_file, kf, self.input_keyframe_path, kf))
        image = Image.open(f'{self.video_file}frames/{kf}.jpg')
        image.show()
        self.input_keyframe.save_graph_data(kf, '  ', '.jpg')

  def selectKeyShot(self, graph, leaflist):
    S = [graph.subgraph(c).copy() for c in nx.connected_components(graph)]
    #KF_list = []
    self.len_shot = int(self.len_shot / len(S))
    for c in range(len(S)):
      central_node = len(S[c].nodes)
      comp_leaf_list = []
      for i in range(central_node):
        if(list(S[c])[i] in leaflist):
          comp_leaf_list.append(list(S[c])[i])
      len_leaf_list = len(comp_leaf_list)
      cn = int(len_leaf_list/2) # find the central node for keyframe strategy

      if not (cn == 0):
        if(len_leaf_list < self.len_shot):
            init_keyshots = 0
            #if(len_shot == 1):
              #end_keyshots = 1
            #else:
            end_keyshots = len_leaf_list - 1
        else:
            init_keyshots = cn - (int(self.len_shot/2))
            end_keyshots = cn + (int(self.len_shot/2))

        if not os.path.isdir(self.input_keyshot_path):
          os.mkdir(self.input_keyshot_path)

        for k in range(init_keyshots, end_keyshots):
            keyshot = str(comp_leaf_list[k]).zfill(6) # save on keyshot the central node
            os.system('cp {}frames/{}.jpg {}{}.jpg'.format(self.video_file, keyshot, self.input_keyshot_path, keyshot))
            self.input_keyshot.save_graph_data(keyshot, '  ', '.jpg') # (path for keyshot, each frame of keyshot, validator to save, extension)

    # Video Generating function
  def generate_video(self):
    image_folder = self.input_keyshot_path[:-1]#'.' # make sure to use your folder

    images = [img for img in os.listdir(image_folder)

            if img.endswith(".jpg") or
              img.endswith(".jpeg") or
              img.endswith("png")]

    # Array images should only consider
    # the image files ignoring others if any

    frame = cv.imread(os.path.join(image_folder, images[0]))

    # setting the frame width, height width
    # the width, height of first image
    height, width, layers = frame.shape
    os.chdir(self.output_skim)

    video = cv.VideoWriter(self.video + '.mp4', 0, self.rate, (width, height))

    # Appending the images to the video one by one
    for image in images:
      video.write(cv.imread(os.path.join(image_folder, image)))

    # Deallocating memories taken for window creation
    cv.destroyAllWindows()
    video.release() # releasing the video generated

  def calc_init(self, i, delta_t, frame_len):
      if((i < delta_t) or delta_t < 0):
          return 0
      elif((i + delta_t) > frame_len):
          return i
      else:
          return i - delta_t

  def calc_end(self, i, delta_t, frame_len):
      if(((i + delta_t) > frame_len) or delta_t < 0):
          return frame_len
      else:
          return i + delta_t


import os
import networkx as nx

class Files:
  def __init__(self, file=' '):
    self.file = file

  def read_graph_file(self, cut_graph_file, cut_graph, cut_number):
    RG = nx.Graph()
    with open(self.file) as f:
      lines = f.readlines()

    cut = len(lines)
    if(cut_graph):
      cut -= (cut_number +1)
    cut_list = []
    for line in lines:
      v1 =int(line.split(", ")[0]) # node 1
      v2 =int(line.split(", ")[1]) # node 2
      w = float(line.split(", ")[2]) # weight

      if(cut >= 0):
        RG.add_edge(v1, v2, weight = w) # include two node and your weight
        cut -=1
        if(cut_graph):
          cut_graph_file.save_graph_data(v1, v2, w)

      else:
        if(w == -1):
          v1 =int(line.split(", ")[0]) # node 1
          v2 =int(line.split(", ")[1]) # node 2
          if(v1 <= v2):
            if(not (v2 in cut_list)):
              RG.add_node(v1)
              cut_graph_file.save_graph_data(v1, ' ', ' ')
              cut_list.append(v2)
              cut -=1
          else:
            if(not (v1 in cut_list)):
              RG.add_node(v2)
              cut_graph_file.save_graph_data(v2, ' ', ' ')
              cut_list.append(v1)
              cut -=1
    return RG

  def create_folder(self, folder_name):
    if not os.path.isdir(f"../models/{folder_name}"):
        os.makedirs(f"../models/{folder_name}")

  def save_graph_data(self, v1, v2, weight):
    f = open(self.file, "a")
    if(v2==' '):
      data = "{}\n".format(v1)
    elif(weight == ".jpg"):
      data = "{}{}\n".format(v1, weight)
    else:
      data = "{}, {}, {}\n".format(v1, v2, weight)
    f.write(data)
    f.close()


from .Summary import Summary
import os
import json
import cv2 as cv
import numpy as np
from datetime import timedelta
from pathlib import Path

class HieTaSumm:
    def __init__(self, **kwargs):
        PACKAGEDIR = Path(__file__).parent.absolute()
        my_file = PACKAGEDIR / 'options.json'

        json_file = open(my_file, 'r')
        default_data = json.loads(json_file.read())
        dataset_videos = kwargs.get('video_path')
        if not dataset_videos:
            dataset_videos = default_data['video_path']
        dataset_frames = kwargs.get('summary_path')
        if not dataset_frames:
            dataset_frames = default_data['summary_path']
        percent = kwargs.get('percent')
        if not percent:
            percent = default_data['percent']
        percent = int(percent)
        alpha = kwargs.get('alpha')
        if not alpha:
            alpha = default_data['alpha']
        alpha = int(alpha)
        rate = kwargs.get('rate')
        if not rate:
            rate = default_data['rate']
        rate = int(rate)
        time = kwargs.get('time')
        if not time:
            time = default_data['time']
        time = int(time)
        hierarchy = kwargs.get('hierarchy')
        if not hierarchy:
            hierarchy = default_data['hierarchy']
        selected_model = kwargs.get('selected_model')
        if not selected_model:
            selected_model = default_data['selected_model']
        is_binary = kwargs.get('is_binary')
        if not is_binary:
            is_binary = default_data['is_binary']
        keyshot = kwargs.get('keyshot')
        if not keyshot:
            keyshot = default_data['keyshot']
        keyframe = kwargs.get('keyframe')
        if not keyframe:
            keyframe = default_data['keyframe']
        
        if not os.path.isdir(dataset_frames):
            os.mkdir(dataset_frames)

        if(os.path.exists(dataset_videos)):
            video_list = os.listdir(dataset_videos)
            video_list.sort() # to guarantee order
            for i in video_list:
                print("------------------------")
                print("{}/{}/".format(dataset_videos, i))
                self.frame_extractor("{}/{}".format(dataset_videos, i), rate, dataset_frames)
        else:
            print(f"do not exist directory or file with path {dataset_videos}")

        if(os.path.exists(dataset_frames)):
            video_list = os.listdir(dataset_frames)
            video_list.sort() # to guarantee order
            for video in video_list:
                if video != '.ipynb_checkpoints':
                    self.hierarchical_summarization(dataset_frames, video, rate, time, percent, alpha, keyshot, keyframe, is_binary, hierarchy, selected_model)

    def frame_extractor(self, video_file, rate, frames):
        # i.e if video of duration 30 seconds, saves 0.5 frame each second = 60 frames saved in total
        SAVING_FRAMES_PER_SECOND = 1/rate
        filename, _ = os.path.splitext(video_file)
        video = frames + "/" + filename.split('/')[-1]
        filename = video + "/frames/"
        if not os.path.isdir(video):
            os.mkdir(video)
        # make a folder by the name of the video file
        if not os.path.isdir(filename):
            os.mkdir(filename)
            # read the video file
            cap = cv.VideoCapture(video_file)
            fps = cap.get(cv.CAP_PROP_FPS)
            saving_frames_per_second = min(fps, SAVING_FRAMES_PER_SECOND)

            saving_frames_durations = self.get_saving_frames_durations(cap, saving_frames_per_second)

            count = 0
            frame_number = 1
            while True:
                is_read, frame = cap.read()
                if not is_read:
                    # break out of the loop if there are no frames to read
                    break
                # get the duration by dividing the frame count by the FPS
                frame_duration = count / fps
                try:
                    # get the earliest duration to save
                    closest_duration = saving_frames_durations[0]
                except IndexError:
                    # the list is empty, all duration frames were saved
                    break
                if frame_duration >= closest_duration:
                    # if closest duration is less than or equals the frame duration,
                    # then save the frame
                    # frame_duration_formatted = formatTimeDelta(timedelta(seconds=frame_duration))
                    number = str(frame_number).zfill(6)
                    frame_number += 1
                    cv.imwrite(os.path.join(filename, "{}.jpg".format(number)), frame) # frame_duration_formatted)), frame)
                    # drop the duration spot from the list, since this duration spot is already saved
                    try:
                        saving_frames_durations.pop(0)
                    except IndexError:
                        pass
                # increment the frame count
                count += 1

    def get_saving_frames_durations(self, cap, saving_fps):
        """A function that returns the list of durations where to save the frames"""
        s = []
        # get the clip duration by dividing number of frames by the number of frames per second
        clip_duration = int(cap.get(cv.CAP_PROP_FRAME_COUNT) / cap.get(cv.CAP_PROP_FPS))
        # use np.arange() to make floating-point steps

        for i in np.arange(0, clip_duration, saving_fps):
            s.append(i)
        return s

    def hierarchical_summarization(self, dataset_frames, video, rate, time, percent, alpha, keyshot, keyframe, is_binary, hierarchy, selected_model):
        Summary(dataset_frames, video, rate, time, hierarchy, selected_model, is_binary, percent, alpha, keyshot, keyframe)

if __name__== "__main__":
    HieTaSumm()


import os
import cv2 
import numpy as np
from .Files import Files

class Frame:
  def __init__(self, model):
    self.model = model

  def load(self, video_file, delta_t, output_graph_file, features_list):
    if(os.path.exists(video_file)):
      frame_list = os.listdir(video_file)
      frame_list.sort() # to garanted the time order

      feat_list_len = len(features_list)
      weight_list = []

      for vertex1 in range(feat_list_len):
        for vertex2 in range(self.calc_init(vertex1, delta_t, feat_list_len),
                                      self.calc_end(vertex1, delta_t, feat_list_len)):
          frame1 = os.path.join(video_file, frame_list[vertex1])
          frame2 = os.path.join(video_file, frame_list[vertex2])
          w = self.compute_similarity_img(frame1, frame2, features_list[vertex1], features_list[vertex2])
          weight_list.append(w)
          output_graph_file.save_graph_data("{}, {}, {:.2f}".format(vertex1, vertex2, w) , ' ', ' ')

  def compute_similarity_img(self, img_path_1, img_path_2, fea_vec_img1, fea_vec_img2): # img_path_1, img_path_2):
      filename1 = os.path.basename(img_path_1).split(".")[0]
      filename2 = os.path.basename(img_path_2).split(".")[0]

      # Compute cosine similarity
      sim_cos = self.calculate_similarity(fea_vec_img1, fea_vec_img2)

      # Read images
      im1 = cv2.resize(cv2.imread(img_path_1), dsize=self.model.img_size_model, interpolation = cv2.INTER_AREA)
      im2 = cv2.resize(cv2.imread(img_path_2), dsize=self.model.img_size_model, interpolation = cv2.INTER_AREA)

      # Concatenate images horizontally
      im12 = cv2.hconcat([im1, im2])

      # Save concatenated image
      folder = Files()
      dst_dir_cos_sim = "../report/cos_sim"
      folder.create_folder(dst_dir_cos_sim)
      dst_dir = f"{dst_dir_cos_sim}/{self.model.name}"
      folder.create_folder(dst_dir)

      new_filename = f"{filename1}_{filename2}"
      cv2.imwrite(f"{dst_dir}/{new_filename}.jpg", im12)

      return sim_cos

  def calculate_similarity(self, vector1, vector2):
      sim_cos = np.linalg.norm(vector1-vector2, 1) * 100 / len(vector2)
      return sim_cos
  
  def calc_init(self, i, delta_t, frame_len):
      if((i < delta_t) or delta_t < 0):
          return 0
      elif((i + delta_t) > frame_len):
          return i
      else:
          return i - delta_t

  def calc_end(self, i, delta_t, frame_len):
      if(((i + delta_t) > frame_len) or delta_t < 0):
          return frame_len
      else:
          return i + delta_t


from setuptools import setup, find_packages

setup(
    name='HieTaSumm',
    version='0.1.1',    
    license='MIT Licence',
    description='A hierarchical approach for video summarization',
    keywords='summarization hierarchy graph',
    author='Leonardo Vilela',
    author_email='leonardo.cardoso.794229@sga.pucminas.br',
    packages=find_packages(),
    package_data={'HieTaSumm': ['options.json']},
    install_requires=[
        'numpy',
        'opencv-python',
        'matplotlib',
        'scikit-learn',
        'torch',
        'Pillow',
        'higra',
        'tensorflow',
        'keras',
        'scipy',
        'networkx',
        'pathlib'
    ],
)